﻿// Copyright 2009 Oliver Sturm <oliver@sturmnet.org> All rights reserved. 

#light

open System
open System.IO

open Helpers

Output.banner "Catching exception"

// catching an exception
try
    // do something that throws an exception
    let foo = 10 / 0
    printfn "%d" foo
with
    | :? InvalidOperationException as e -> printf "InvalidOperationException: %s" e.Message
    | e -> printfn "Other exception: %s" e.Message

// this works also in expressions
let result = 
    try
        Some(10 / 0)
    with
        | _ -> None
        
Output.banner "A custom exception"

// custom exceptions
exception MyOwnException of string * int

// raising and catching a custom exception
try
    raise (MyOwnException("An error has occurred", 42))
    // These throw a FailureException
    // failwith "There's a problem"
    // failwithf "There's a problem with a number: %d" 42
with
    | MyOwnException(strVal, intVal) -> 
        printfn "Got exception with str='%s', int=%d" strVal intVal
        // rethrow the exception
        // rethrow()

Output.banner "Try/finally"

// try/finally
try
    try
        failwith "test"
    finally
        printfn "Finally block"
with 
    | e -> printfn "Caught"
    
// working with IDisposable objects

Output.banner "IDisposable"

// using calls Dispose on the disposable object that is 
// created by the expression in its first parameter,
// after the function given as the second parameter ends
let writeFile() = 
    using (File.CreateText("file.txt"))
        (fun file -> 
            file.WriteLine("Stuff in the file"))

writeFile()

// alternatively, use calls IDisposable.Dispose when the block ends
let writeFile2() = 
    use file = File.CreateText("file2.txt")
    file.WriteLine("Stuff in other file")

writeFile2()

// ---------------------- snip (demo)
    
Output.banner "Casting"
    
// casting up
let obj1boxed = box 1
let objTextCast = "Text" :> obj

// casting down
let int1 = obj1boxed :?> int

// this results in an exception
try
    let str = obj1boxed :?> string
    printfn "The string is '%s'" str
with 
    | _ as e -> printfn "Caught exception, message is '%s'" (e.Message)

Output.banner "Type checking in match expressions"

// checking types in match expressions
let checkThis (x:obj) = 
    match x with 
    | :? int as i -> printfn "This is the int %d" i
    | :? string as s -> printfn "This is the string '%s'" s
    | _ -> printfn "I don't have a clue what this is"

checkThis obj1boxed
checkThis objTextCast

// match expressions can also be used in assignments
let myValue = 
    match objTextCast with 
    | :? string as s -> s
    | _ -> "dummy string"
    
// gives int
let i = 10

// convert to int64
// convert functions exist for all standard numeric data types
let i64 = int64 i

// ---------------------- snip (demo)
    

Output.banner "Lists"

let outputList list = 
    for x in list do printf "%A " x
    printf "\n"
    
// simple lists
let intList = [10;20;30]
let stringList = List.empty<string>
let generatedList = List.init 10 (fun x -> x * x)
let anotherGeneratedList = [ for i in 1..5 -> i * i ]

outputList intList
outputList stringList
outputList generatedList

// head and tail

printfn "%A" (List.hd intList)
outputList (List.tl intList)

let rec listLength list = 
    match list with 
    | [] -> 0
    | hd :: tl -> 1 + (listLength tl)

printfn "length of intList: %d" (listLength intList)
printfn "length of generatedList: %d" (listLength generatedList)

// take the first n elements from list l
let rec takeFromList n l =
    match n, l with
    | 0, _ -> []
    | _, [] -> []
    | _,  (hd :: tl) -> hd :: takeFromList (n - 1) tl
    

outputList (takeFromList 3 [1;3;5;7;11;13;17;19])

// Sequences

Output.banner "Sequences"

let squarePairs = seq { for x in 0 .. 20 -> (x, x*x) }
let squarePairsMaxVerbose max = 
    seq { 
        for x in 0 .. max -> 
            printfn "Calculating next value"
            (x, x*x) 
    }
outputList squarePairs
outputList (squarePairsMaxVerbose 10)

let createOddNumbers max = { 1 .. 2 .. max}
outputList (createOddNumbers 20)

let explicitSequence = 
    seq {
        yield "This"
        yield "That"
        yield "Something else"
        yield! ["A"; "nested"; "list"]
    }
    
outputList explicitSequence

let otherExplicitSequence = 
    seq {
        let x = ref 3
        while true do
            yield !x
            x := !x + 1
    }
    
outputList (Seq.take 5 otherExplicitSequence)
