﻿// Copyright 2009 Oliver Sturm <oliver@sturmnet.org> All rights reserved. 

#light

open System 

open Helpers

Output.banner "Working with data"

let outputList list = 
    for x in list do printfn "%A " x
    printf "\n"
    
// Each customer has a name, an age, a city and a sales volume
let customers =
    [
        ("Bobby", 32, "Bristol", 2075.87M);
        ("Harry", 37, "Birmingham", 122.54M);
        ("Karen", 45, "London", 324.78M);
        ("Lisa", 19, "Manchester", 837.00M);
        ("Karl", 28, "London", 16.88M);
        ("Susan", 25, "Dover", 97.13M)
    ]                

// filtering - Where in LINQ
let customersOver30 = customers |> Seq.filter (fun (_, age, _, _) -> age > 30)
outputList customersOver30

// mapping - Select in LINQ
let goodCustomerNames = 
    customers |> Seq.filter (fun (_, _, _, salesVolume) -> salesVolume > 300M)
              |> Seq.map (fun (name, _, _, _) -> name)

outputList goodCustomerNames

let customersWithK = 
    customers |> Seq.map (fun (name, _, _, _) -> name)
              |> Seq.filter (fun name -> name.StartsWith "K")

outputList customersWithK

// aggregating - Aggregate in LINQ
let totalSalesVolume = 
    customers |> Seq.map (fun (_, _, _, salesVolume) -> salesVolume)
              |> Seq.fold (fun r x -> r + x) 0M 

printfn "%M" totalSalesVolume

// it is sometimes possible to use operators as the 
// functions for fold:
let sumUpList l = Seq.fold (+) 0 l

Output.banner "Function construction"

// combining partial application and standard higher order functions
// to create new, more specific functions
let intSequenceAdder : seq<int>->int = Seq.fold (+) 0

// This results in a Value Restriction error, because 
// when partial application is combined with the type
// inference process and thus the result of the partial
// application is more restricted than the original function,
// F# wants "confirmation of the intention", either in 
// the form of type restrictions (above) or by making the
// whole construct a syntactic function instead (below).
//
// let squareMapper = Seq.map (fun x -> x * x)

let squareMapper l = Seq.map (fun x -> x * x) l

let testList = [10;20;30]
printfn "%d" (intSequenceAdder testList)
outputList (squareMapper testList)

// A function for prime calculation based on fold
let isPrime x = 
    Seq.fold (fun p v -> p && (x % v <> 0)) true {2 .. (x/2)}

printfn "%b" (isPrime 13)
printfn "%b" (isPrime 10)
printfn "%b" (isPrime 47)

Output.banner "Map and filter as applications of fold"

// fold is the most powerful of the three (map, filter, reduce)
// The following functions show that map and filter can be implemented
// on the basis of fold - not very performant implementations
// and they also use lists for simplicity, so not optimal for 
// use with sequences. Just to prove the point!

// map as an application of fold
let map f l = 
    Seq.fold (fun r x -> r @ [f x]) [] l

outputList (map (fun x -> x * x) [1;2;3])

// filter as an application of fold
let filter f l = 
    Seq.fold (fun r x -> r @ if f x then [x] else []) [] l

outputList (filter (fun x -> x > 3) [1;3;5;9])
