﻿// Copyright 2009 Oliver Sturm <oliver@sturmnet.org> All rights reserved. 

#light

open System
open Helpers

Output.banner "Simple immutable types"

// Simple immutable types

// Members and an instance method
type Point = 
    { X: float; Y: float }
    member x.Shift(dx, dy) = { x with X = x.X + dx; Y = x.Y + dy }

// A static method (Circle). Note that Area is not a function!
type Ellipse = 
    { RadiusX: float; RadiusY: float }
    member x.Area = 
        printfn "Calculating Area"
        System.Math.PI * x.RadiusX * x.RadiusY
    static member Circle(radius) = { RadiusX = radius; RadiusY = radius }

// construction

let p1 = { X = 10.3; Y = 11.4 }
printfn "%A" p1
let p2 = p1.Shift(3.0, 3.0)
printfn "%A" p2

// using an explicit type in case of conflicts
let ellipse : Ellipse = { RadiusX = 3.0; RadiusY = 5.4 }
printfn "%A" ellipse
printfn "Area: %f" ellipse.Area
let circle = Ellipse.Circle(10.3)
printfn "%A" circle
printfn "Area: %f" circle.Area

Output.banner "Interface created by object expression"

// an interface
type IInOut = 
    abstract Input: unit -> string
    abstract Output: obj -> unit

// a function that uses an interface
let greet (io: IInOut) = 
    io.Output "What's your name?"
    let name = io.Input()
    io.Output ("Hi " + name)

// a call to a function that uses an interface, creating
// an interface implementation using an object creation expression
greet { new IInOut with 
        member x.Input() = Console.ReadLine()
        member x.Output(o) = printfn "%A" o }

Output.banner "Constructed types"
        
// a constructed type with an additional ctor
type Rectangle(height: float, width: float) = 
    let diagonal = sqrt(height * height + width * width)
    do if diagonal > 10.0 then failwith ("this rect is too large: " + diagonal.ToString())
    member x.Height = height
    member x.Width = width
    member x.Diagonal = diagonal
    new(lengthofside) = Rectangle(lengthofside, lengthofside)
    
// instantiation
let rect1 = Rectangle(6.0, 4.7)
printfn "rect1 diagonal: %f" rect1.Diagonal
let square = Rectangle(4.5)
printfn "square diagonal: %f" square.Diagonal

Output.banner "Constructed type with interface"


// this constructed type implements an interface
type GreeterIO() = 
    interface IInOut with
        member x.Input() = Console.ReadLine()
        member x.Output(o) = printfn "%A" o

// It can be used with the greet function
// To get to the interface, an explicit upcast is necessary (:>)
greet (GreeterIO() :> IInOut)

Output.banner "Dynamic object expression"

// object expressions can be dynamic
let createGreeterIO input output = 
    { new IInOut with
        member x.Input() = input()
        member x.Output(o) = output(o) }

// create a greeter implementation using particular functions
// the output function uses partial application to match the signature
let myGreeter = createGreeterIO Console.ReadLine (printfn "%A")

greet myGreeter

Output.banner "Partially implemented class types"

// an abstract base class
[<AbstractClass>]
type GreeterBase() =
    abstract Input : unit -> string
    abstract Output : obj -> unit
    abstract FormatInput : string -> string
    abstract FormatOutput : string -> string
    default x.FormatInput(s) = s
    default x.FormatOutput(s) = s
    member x.Greet() = 
        x.Output (x.FormatOutput("What's your name?"))
        let name =  x.FormatInput(x.Input())
        x.Output (x.FormatOutput("Hi " + name))
    
// a derived class giving default implementations 
// for some abstract members
type ConsoleGreeter() = 
    inherit GreeterBase()
    default x.Input() = Console.ReadLine()
    default x.Output(o) = printfn "%A" o
    
// using an object expression to change an abstract member
// composition instead of inheritance
type HTMLGreeter() = 
    let greeter = 
        { new ConsoleGreeter() with
            member x.FormatOutput(s) = "<b>" + s + "</b>" }
    member x.GreetHTML() = 
        printfn "<h2>Great greeting sample coming up!</h2>"
        greeter.Greet()

let htmlGreeter = HTMLGreeter()
htmlGreeter.GreetHTML()

Output.banner "Augmenting existing types"

// opening an existing type
type Int32 with 
    member x.SpecialOutput() = printfn "Value is: %d" x

let intVal = 10
intVal.SpecialOutput()

type Thing =
    | OneThing of string
    | OtherThing of int

type Thing with 
    member x.OutputThing() = 
        match x with 
        | OneThing(s) -> printfn "Thing with a string: %s" s
        | OtherThing(i) -> printfn "An int thing: %d" i

let cgreeter = ConsoleGreeter()
// at this point, this call is not possible
// cgreeter.AdditionalMember()
        
type ConsoleGreeter with
    member x.AdditionalMember() = printfn "Nothing"

cgreeter.AdditionalMember()

// mutually dependent types

type TypeA() = 
    let b = TypeB()

and TypeB() =
    let a = TypeA()

// mutually dependent functions

let rec functionA() = functionB()
and functionB() = functionA()

