﻿// Copyright 2009 Oliver Sturm <oliver@sturmnet.org> All rights reserved. 

#light

open System
open Helpers 

Output.banner "Mutable properties"

// basic class with two mutable properties - F# understands the 
// apparently conflicting names exactly right
type Person(firstName: string, lastName: string) =  
    let mutable firstName = firstName
    let mutable lastName = lastName
    
    member p.FirstName with get() = firstName and set(n) = firstName <- n
    member p.LastName with get() = lastName and set(n) = lastName <- n

    default p.ToString() = String.Format("Person {0} {1}", firstName, lastName)
    
let p = Person("Anna", "White")

// Optionally, arguments can be named
let p2 = Person(firstName = "Anna", lastName = "White")

Console.WriteLine p

p.FirstName <- "Billy"
p.LastName <- "Bott"

Console.WriteLine p

Output.banner "Mutable arguments"

// Arguments can also be optional
type OtherPerson(?firstName:string, ?lastName:string) =
    let firstName = defaultArg firstName ""
    let lastName = 
        match lastName with
        | None -> "Smith" // everybody who doesn't have a last name is called Smith
        | Some n -> n
    member x.FirstName = firstName
    member x.LastName = lastName

    // optional arguments are also available on member functions
    member x.DoSomething(?value:int) = 
        if value.IsSome then printfn "Hey, got a value"
        
    default p.ToString() = String.Format("Person {0} {1}", firstName, lastName)

// Harry Smith
let op1 = OtherPerson("Harry")
Console.WriteLine op1

// some unnamed Jones
let op2 = OtherPerson(lastName = "Jones")
Console.WriteLine op2

// an unnamed Smith
let op3 = OtherPerson()
Console.WriteLine op3

Output.banner "Implementation inheritance"

// a derived type
type Worker(firstName:string, lastName:string, salary:decimal) = 
    inherit Person(firstName, lastName)
    member x.Salary = salary
    
    // accessing base class functionality
    default x.ToString() = String.Format("{0} (Salary: {1})", base.ToString(), salary)

let w1 = Worker("Jenny", "Myers", 1500m)
Console.WriteLine w1

type Manager(firstName:string, lastName:string) =
    inherit Person(firstName, lastName)
    
    let mutable workers = List.empty<Worker>
    
    // This is an indexer property
    member x.Item 
        with get(index) : Worker option = 
            try
                Some(List.nth workers index)
            with
                | _ -> None
    
    member x.AddWorker w = 
        workers <- List.append workers [w]

let m = Manager("Peter", "Boyd")

let worker1 = m.[0] // exception, so I get None
Output.outputOption worker1

m.AddWorker (Worker("Suzy", "Carlsson", 1500m))
m.AddWorker (Worker("Jenny", "Myers", 1600m))
m.AddWorker (Worker("Harry", "Jones", 1500m))

// Get back Jenny Myers
let worker2 = m.[1] 
Output.outputOption worker2

// method overloading

Output.banner "Method overloading"

type FlexibleWalker() = 
    // This does *not* work, of course
    //member x.Walk x y = 
    //    printfn "Walking with %d and %d" x y
    //member x.Walk x = 
    //    printfn "Walking with %d" x

    // overloading by number of arguments
    member x.Walk(a, b) = 
        printfn "Walking with %d and %d" a b
    member x.Walk(a) = 
        printfn "Walking with %d" a
        
    [<OverloadID("WalkWithString")>]
    member x.Walk(s) = 
        printfn "Walking with the string '%s'" s    
    
    member x.Walk() = 42
    [<OverloadID("WalkReturningString")>]
    member x.Walk() = "A String"
    
    
let fw = FlexibleWalker()
// tuple syntax needed!
fw.Walk(10, 20)
fw.Walk(10)
fw.Walk("Somebody")

// accessibility specifiers

// a private type
type private PrivateClass() = 
    member x.DoSomething() = ignore

// a private function
let private DoSomething() = ignore

type MyClass() = 
    // a private member
    member private x.DoSomething() = ignore

type MyBaseClass() = 
    member x.Go() = ignore
    
    abstract GetValue: unit -> int
    default x.GetValue() = 10

type MyDerivedClass() = 
    inherit MyBaseClass()
    // this doesn't work because Go isn't abstract (virtual)
    //override x.Go() = ignore
    
    // "override" and "default" mean exactly the same thing
    override x.GetValue() = 42
